
#include <Arduino.h>
#include "settings.h"
#include <HB_LiquidCrystalMod.h>
#include <HB_OneWire.h>
#include <HB_tempsensor.h>
#include <HB_encoder.h>
#include <HB_eeprom.h>
#include <HB_batterie.h>
#include <HB_tone.h>
#include <HB_SoftPWM.h>
#include <avr/pgmspace.h>
#include "icons.h"
#include "woerter.h"
#include <avr/power.h>
#include <HB_base64.h>

/**
 * Struct t_sensordata 
 * speichert alle aktuellen Sensordaten
 */
struct t_sensordata{
 float temperatur;        //ermittelte Temperatur in Grad Celsius
 unsigned int verbrauch;  //verbrauchte Kapazität
 float spannung;          //aktuelle Spannung in Volt
 uint8_t altspannung;     //zuvor ermittelte Spannung in Volt (zum Ermitteln ob sich etwas geändert hat un PWM angepasst werden muss)
 uint8_t prozent;         //Akkuladestand in %
} sensordaten;


_tSettings_  hbsettings, hbtempsettings;




boolean setzeStandartWerte = false;
boolean setzeVorherigeWerte = false;
boolean speichereAktuelleWerte = false;
boolean zeigeBestaetigung = false;
boolean verbrauchZuruecksetzen = false;
boolean starteCountdown = false;
boolean countdownMitHeizen = false;
boolean warnschwelleErreicht = false;
boolean warnschwellenMerker = false;            //Wird TRUE, wenn die Warnschwellenmeldung erstmalig dem Nutzer gezeigt wird (verhintert mehrfaches Auslösen von z.B. LEDBLINK)
boolean EncoderButtonPressed = false;           //Merkt sich, ob der Encoder gedrückt wird oder nicht (wird true wenn Event auf "Press" und false bei Event auf "release"
boolean stringComplete = false;                 //wird TRUE wenn festgestellt wurde, dass der aktuelle Befehl vollständig empfangen wurde
boolean akustischerFehlerTeil=true;             //das Fehlerpiepsen besteht aus zwei Tönen. Hier merken wir, welchen der beiden Töne wir spielen sollen
boolean letzterAkustischerFehlerTeil = false;
boolean externeVersorgung = false;
boolean fehlerUnbestaetigt = false;              //Wird beim "Quittieren" der aufgetretenen Fehler false. Wird wieder TRUE wenn ein neuer Fehler auftritt.
boolean fetWarnungMerker = false;               //Wird TRUE wenn zum ersten mal die FET-Warnung ausgelöst wird (wird false wenn danach alles wieder ok)
boolean warnschwelleUnbestaetigt = true;


unsigned long sensorIntervall = 0;              //merkt den letzten Intervall für die Sensoren
unsigned long encoderPressed = 0;               //Wann wurde der Encoder gedrückt
unsigned long bestaetigungsAnzeige = 0;         //Wann wurde die Bestätigungsanzeige aktiviert?
unsigned long bildschirmWechsel = 0;            //Wann wurde zwischen Bildschirmdarstellung 1 und 2 gewechselt
unsigned long heizungIntervall = 0;             //Wann wurde die Heizung das letzte Mal ein oder ausgeschaltet
unsigned long ventilatorIntervall = 0;          //Wann wurde der Ventilator das letzte Mal ein oder ausgeschaltet
unsigned long LEDZielZeit = 0;                  //Merkt sich, wann die LED ein oder ausgeschaltet  WERDEN SOLL 
unsigned long bildschirmFehlerWechsel = 0;      //Wann wurde zwischen Fehlermelsung Teil1 und Teil2 gewechselt
unsigned long countdownStart = 0;               //Wann wurde der Countdown gestartet
unsigned long letzterDreh=0;                    //Wann wurde zuletzt der Encoder gedreht? - Nötig um zwischen langsamen und schnellen Drehen zu unterscheiden
unsigned long letzerAkustischerFehler=0;        //Wann wurde das letzte Mal ein Fehler-Piepsen ausgegeben
unsigned long letzterEEPROMSpeicher =0;         //Wann wurde das letzte mal die aktuelle Kapazität in das EEPROM abgespeichert

#ifndef D4ZEILEN  /*Beim 4-Zeilen-Display gibt es keine zweite Seite*/
uint8_t displayPage=DISPLAY_MAIN;               //Welche "Startseite" wird gerade angezeigt
uint8_t vorherigeDisplayPage=DISPLAY_MAIN;
#endif

uint8_t countdownZeitStunden = 0;               //Countdown-Zeit
uint8_t countdownZeitMinuten = 0;               //Countdown-Zeit
uint8_t countdownZeitSekunden = 0;              //Countdown-Zeit
uint8_t heizungCode = HEIZUNG_AUS;              //Aktueller Betriebszustand der Heizung
uint8_t ventilatorCode = VENTILATOR_AUS;        //Aktueller Betriebszustand der Heizung
uint8_t LEDCode = LED_AUS;                      //Aktueller Betriebszustand der Display-LED
uint8_t intervallZaehlerTSensor = 0;            //Speichert die Anzahl der Intervalle seit letzter Temperaturabfrage. Die T-Bibliothek arbeitet leider mit vielen Delays. Wir fragen sie deswegen nur alle paar Intervalle ab. 
uint8_t intervallZaehlerDisplay =0;
uint8_t state = STATE_NORMAL;                   //Zustandsautomat: aktueller Zustand;
uint8_t state_afterConfirm = STATE_INVALID;              //Soll nach dem visuellen BEstätigen einer Aktion ein neuer Status angenommen werden wird dieeser hier abgelegt und nach Beenden Der Bestätigungsanzeige geladen
uint8_t fehlerBitPosition = 0;                  //Welcher Fehler aus dem Fundus aller aufgetretenen Fehler wird GERADE angezeigt
uint8_t fehlerAnzeige = 0;                      //0 = "X Fehler"    1 = "  X von Y  "     2 = eigentliche FehlerMeldung
uint8_t fehlerXvonY = 1;                        //Der Wievielte Fehler aus allen aufgetretenen wird gerade angezeigt


uint16_t fehlerCode = ERROR_NOERROR;            //Speichert, WO ein Fehler aufgetreten ist
uint16_t alterFehlerCode = ERROR_NOERROR;       //zum Vergleich, ob neue Fehler hinzugekommen sind
uint16_t aktuellAngezeigterFehler = ERROR_NOERROR;

float unterSpannungsGrenze = 0.0f;              //merken wir uns hier noch mal um nur eine Adresse an die Batterie-Bibliothek übergeben zu müssen. Ein variables konstrukt aus &array[&adresse1][&adrsse2] war mir nicht möglich  -  verbesserungswürdig.
#ifdef ENABLEPCCOM
  String inputString = "";                        //Speichert die empfangenden Daten
#endif
byte encoderSignal=0;                           //Speichert den von der Encoder-ISR gelieferten Encoder-Zustand. 0 = kein Zustand (letzter Zustand abgearbeitet)   
char branding[17];                              //Hierin wird das Branding gespeichert

#include "menusteuerung.h"


EepromHandler* eeprom;
Encoder*       encoder;
TempSensor*    tsensor;
Batterie*      batterie;
HBTone*        hbtone;
LiquidCrystal lcd(12, 11, 8, 7, 6, 5);





/**--------------------------------------------------------------------------------------------------------------------------------
 * Setup-Routine
 *
 */
void setup() {
  
  power_twi_disable();
  //power_usart0_disable();
  power_spi_disable();
   
   
  Serial.begin(115200);
  /*Reserviert etwas Speicher für die Serielle Kommunikation*/
  #ifdef ENABLEPCCOM
    inputString.reserve(10);
  #endif
  
  //prüfe, ob wir auf interner oder externer Versorgung laufen
  pinMode(pinIntExt,INPUT);
  if(digitalRead(pinIntExt)==HIGH) externeVersorgung = true;
  #ifdef FORCE_EXTERN
    externeVersorgung = true;
  #endif 
 
 
  
  //Instanzen der Klassen unserer Bibliotheken erzeugen
  encoder = new Encoder(&encoderCallback);
  eeprom = new EepromHandler();
  tsensor = new TempSensor(&ErrorCallBack, pinTSensor, TSensorAufloesung);
  batterie = new Batterie(&ErrorCallBack);                                             
  hbtone = new HBTone();
 

  //Initialisiert die TempSensor-Bibliothek.
  tsensor->init();
  tsensor->getTemp(&sensordaten.temperatur);
  
 
  //LCD Initialisieren
  #ifdef DISPLAY4
    lcd.begin(4,16);                // Objekt erstellen. Parameter: 4 Zeilen, 16 Zeichen je Zeile
  #else
    lcd.begin(2,16);                // Objekt erstellen. Parameter: 2 Zeilen, 16 Zeichen je Zeile
  #endif
  
    //SoftPWM initialisieren
  SoftPWMBegin();
  

  
  //Icons in den Speicher des Displaycontrollers laden
  ladeIcons();
  
  //Läd gemeinsam verwendete Wörter
  ladeGemeinsameWoerter();
  
  
  //Feststellen ob Bereits ein Setting im EEPROM liegt (und dieses laden)
  //oder ob Standard-Werte geladen werden müssen.
  //Die Methode gibt TRUE zurück wenn Standart-Werte geladen wurden, so zeigen wir dann eine entsprechende Meldung im Display an
  if(ladeEEPROMWerte()){
    lcd.setCursor(0,0);
    lcd.print(F("Wegen Update"));
    lcd.print(leer16Zeichen);
    lcd.setCursor(0,1);
    lcd.print(F("EEPROM loeschen"));
    lcd.print(leer16Zeichen);
    delay(2000);
    lcd.setCursor(0,0);
    lcd.print(F("Standartwerte"));
    lcd.print(leer16Zeichen);
    lcd.setCursor(0,1);
    lcd.print(F("geladen"));
    lcd.print(leer16Zeichen);
    delay(2000);
    lcd.setCursor(0,0);
    lcd.print(leer16Zeichen);
    lcd.setCursor(0,1);
    lcd.print(leer16Zeichen);
  };
  
  
  //Wenn Verbräuche gespeichert werden sollten, dann laden wir den zuletzt gespeicherten Wert
  if(!externeVersorgung && hbsettings.sollSpeichern) ladeLetztenVerbrauch();
  
  
  //initialisiert die EEPROM-Bibliothek !!muss nach dem Laden der letzten Kapazität durchgeführt werden, wählt neuen Slot aus!!!
  eeprom->init();
  
  //ließt das Branding aus dem EEPROM
  eeprom->getBranding(branding);
  
  boolean notBranded = true;
  //prüfe das Branding
  uint16_t checksum = 0;
  for(uint8_t i=0; i< strlen(branding); i++) {
     checksum+= (uint8_t)branding[i];
  }
  checksum &= 0xFF;  //Niederwertigestes Byte wird nur betrachtet
  checksum ^= 0xFF;  //Zweierkomplement bilden (XOR)
  checksum += 1;

  for(uint8_t i=0; i < sizeof(erlaubeBrandingPruefsummen);i++) {
    if(checksum == erlaubeBrandingPruefsummen[i]) notBranded = false;
    
  }
  
  
  if(notBranded) { 
    hbtone->play(1); 
    eeprom->setBranding(defaultBranding);
    eeprom->getBranding(branding);
    delay(1000);
  };
  
  /*Instanz der Batteriebibliothek mit Initialwerten befüllen*/
  batterie->setWarnschwelle(externeVersorgung?NULL:&hbsettings.warnschwelle);
  batterie->setAbschaltgrenze(externeVersorgung?NULL:&hbsettings.abschaltgrenze);
  batterie->setUnterspannungsGrenze(externeVersorgung?NULL:&unterSpannungsGrenze);
  batterie->setMaximalspannung(maxSpannung);
  batterie->setNominalspannung(nomSpannung);
  batterie->setMaximalstrom(maxStrom);
  batterie->setNominalstrom(nomStrom);
  batterie->setPinSpannung(pinUSensor);
  batterie->setPinStrom(pinISensor);
  batterie->setSchwankungsDelay(spannungSchwankungVerzoegerung);
  batterie->setKalibrierwert(hbsettings.kalibrierWert);
  batterie->setKapazitaet(hbsettings.kapazitaet);  //Wurde bereits eine verbrauchte Kapazität gespeichert, dann stellen wir diese ein


  
  //Menüstruktur erstellen
  setupMenu();
  
  //Aktiviere die LED-Hintergrundbeleuchtung des Displays für eine Weile
  //pinMode(pin_backlight,OUTPUT);
  LEDSteuerung(LED_AN);
  
  //Begrüßung und Versionsnummer anzeigen
  lcd.setCursor(1,0);
  lcd.print(F("HeatBox V"));
  lcd.print(hbVersionString);
  lcd.setCursor(0,1);
  lcd.print(branding);
  
  //Ton spielen
  hbtone->play(2);
  
#ifndef D4ZEILEN  
  //Beim Starten wird die ISP des Encoders ein mal aufgerufen (Grund unbekannt). Dadurch wird displayPage bereits ein mal verändert. Wir setzen es hier noch mal auf 0
  displayPage=DISPLAY_MAIN;
#endif

  //Anzeige eine Weile lang stehen lassen
  delay(2500);
  
  //Anzeige löschen
  lcd.clear();
}


/**--------------------------------------------------------------------------------------------------------------------------------
 * Loop-Routine
 *
 */
void loop(){
  
 #ifdef ENABLEPCCOM
  //Wertet die empfangenen Daten aus wenn das Zeilenende erkannt wurde
  if (stringComplete) {  handleSerialInput(); }
 #endif
  
  //Hat die Encoder-ISR einen neuen Zustand gemeldet rufen wir die entsprechende Behandlungsroutine auf.
  if(encoderSignal>0) handleEncoderFlag();
  
  //Die Sensoren werden nur in bestimmten Intervallen abgefragt... 
  //prüfen, ob wieder ein Intervallbegin erreicht wurde
  if(sensorIntervall+defIntervallSensoren < millis()) {
      sensorIntervall=millis();
    
      //die T-Bibliothek arbeitet leider mit vielen Delays. Wir fragen sie deswegen nur alle paar Intervalle ab
      //Also prüfen wir, ob die Bibliothek wieder ein mal abgefragt werden muss.
      if(++intervallZaehlerTSensor == multiplikatorTSensor) {
          tsensor->getTemp(&sensordaten.temperatur);
          intervallZaehlerTSensor=0;
      }
      batterie->tick();        //Die Batterie-Bibliothek sollte so oft wie Möglich angetickt werden damit die Messungen möglichst genau sind
    
      sensordaten.spannung =  batterie->getSpannung();
      sensordaten.verbrauch = batterie->getVerbrauch();
      sensordaten.prozent =   batterie->getProzentRest();
    
      /*Überschreitet der zuletzt gemessene Strom eine bestimmte Schwelle obwohl nicht gezeizt wird soll ein Fehler ausgelöst werden*/
      if((heizungCode!=HEIZUNG_AN) && (batterie->getAktuellerStrom() > stromSchwellePause)) {
          if(!fetWarnungMerker) {
              fetWarnungMerker=true;
          } else {
              ErrorCallBack(NULL,ERROR_ENDSTUFE);
          }       
      } else {
          if(fetWarnungMerker) fetWarnungMerker=false;
      }
    
      //Wenn die Kapazität in den EEPROM abgespeichert werden soll und es wieder an der Zeit ist...
      if(!externeVersorgung && hbsettings.sollSpeichern && (letzterEEPROMSpeicher+intervallEEPROMWrite < millis())) {
          eeprom->writeCapacity(&sensordaten.verbrauch);
          letzterEEPROMSpeicher = millis();
      }
    
      if(++intervallZaehlerDisplay == multiplikatorDisplay) {
          intervallZaehlerDisplay=0;
      
          switch(state) {
              case STATE_MENU: break;
              case STATE_MENU_CHANGEVALUE: break;
              case STATE_COUNTDOWN: zeigeCountdown(); break; 
              default: zeigeHauptschirm();
          };
      } 


      /* Füge in einige Menüpunkte noch weitere Daten ein*/
  
      if((state == STATE_MENU || state == STATE_MENU_CHANGEVALUE) && currentItem == &sonst4) {
          lcd.setCursor(8,1);
          if(heizungCode == HEIZUNG_AN)
              lcd.print("H ");
          else
              lcd.print("0 ");
          lcd.print(padspace(sensordaten.spannung,2,5));
          lcd.print(F("V "));
      }
    } //if(sensorIntervall+defIntervallSensoren < millis())

  
  /*
  Für manche Aktionen wollen wir dem Nutzer eine Kurze Bestätigung geben.
  Diese Bestätigung wird eine Kurze Zeit lang im Display angezeigt.
  Um nicht mit Delay zu arbeiten (und in den ISR's geht Delay sowieso nicht)
  nutzen wir einfach die abgelaufene Zeit.
  Anschließend werden die Bestätigung durch die aktuelle Wertezeile ersetzt
  */
  if(zeigeBestaetigung && (bestaetigungsAnzeige+bestaetigungsDauer)<millis()){
      zeigeBestaetigung = false;
      if(state_afterConfirm == STATE_INVALID) { zeigeMenuWerte(false);}  //Soll kein neuer State angenommen werden, dann zeige einfach wieder die Werte des aktuellen Menüs
      else {state= state_afterConfirm; lcd.clear();}                      //Ansonsten lade den neuen State  (wurde eingeführt um nach dem Speichern in die Hauptübersicht zurück zu kommen)
  }
  
  /*
  Heizung und Lüfter wollen je nach Status oder in bestimmten Zeitabständen (je nach Nutzervorgabe)
  ein bzw. ausgeschaltet werden.
  Hierzu werten wir den <i>fehlerCode</i> sowie die eingestellten Intervalle aus und setzen den 
  Betriebszustand von Heizungim Bytes <i>heizungCode</i> und den Zustand des Ventilators
  im Flag <i>ventilatorAn</i>
  */
  
  if(fehlerCode!=alterFehlerCode) {
     #ifndef D4ZEILEN
       displayPage=DISPLAY_SECONDARY;
     #endif
     heizungCode = HEIZUNG_AUS;
     ventilatorCode = VENTILATOR_AUS;
     LEDSteuerung(LED_BLINK);
     heizungSteuerung();
     ventilatorSteuerung();
     alterFehlerCode = fehlerCode;
     #ifdef HBDEBUG
       Serial.println(F("Heizung+Ventilator AUS wegen Fehler"));
     #endif
  }
  
 if((fehlerCode!=ERROR_NOERROR)) {
      if(hbsettings.alarmTon && fehlerUnbestaetigt) {
          if ( (letzerAkustischerFehler+ETON_PAUSE) < millis()) {
              hbtone->play(akustischerFehlerTeil?3:4);
              akustischerFehlerTeil=!akustischerFehlerTeil;
              letzerAkustischerFehler=millis();
          }
      } // if hbsettings.alarmTon
  } else {
    
    if(warnschwelleErreicht && !warnschwellenMerker) {
        LEDSteuerung(LED_BLINK);
        warnschwellenMerker=true;    //Merken, dass wir die Warnschwelle bereits ein mal ausgewertet haben
    }
    if(hbsettings.alarmTon && warnschwelleErreicht  && warnschwelleUnbestaetigt) {
        if ( (letzerAkustischerFehler+ETON_PAUSE) < millis()) {
            hbtone->play(akustischerFehlerTeil?3:4);
            akustischerFehlerTeil=!akustischerFehlerTeil;
            letzerAkustischerFehler=millis();
        }
    } // if hbsettings.alarmTon 
  
  
    /*
    COUNTDOWN
    Wir zählen die voreingestellte Zeit herunter bis sie bei 0 ist und wechseln
    dann in den normalen Betriebszustand  
    */
    if(state==STATE_COUNTDOWN) {
        while((millis()-countdownStart)>1000) {
            if(countdownZeitSekunden>0) {
                countdownZeitSekunden--;
                countdownStart+=1000;
            } else if(countdownZeitMinuten>0) {
                countdownZeitSekunden=59;
                countdownZeitMinuten--;
            } else if(countdownZeitStunden>0){
                countdownZeitStunden--;
                countdownZeitMinuten=59;
            }; 
            if((countdownZeitSekunden + countdownZeitMinuten + countdownZeitStunden)==0) {
                state=STATE_NORMAL;
                lcd.clear();
                hbtone->play(2);
            }
       }; //while
    } // if(state==STATE_COUNTDOWN)
   
    // else {   --- wieder entfernt weil jetzt im CountDown auch eine Warmhaltefunktion existiert
    
      /*
      HEIZUNG:
      Wurde die Zieltemperatur erreicht gehen wir in den Wartezustand und hoffen, dass dieser
      lange Zeit anhält :) (dann haben wir einen guten Koffer gebaut).
      */
      uint8_t temp=0;
      if(state==STATE_COUNTDOWN && countdownMitHeizen) temp = hbsettings.warmhalteTemperatur; 
      else temp = hbsettings.zieltemperatur;
      if(sensordaten.temperatur >= temp) {
        heizungCode = HEIZUNG_WARTET;
        heizungSteuerung();
      }
    
      else 
    
      /*
      HEIZUNG:
      Wir befinden uns im Wartezustand, da vormals die Zieltemperatur erreicht wurde.
      Wir prüfen, ob die aktuelle Temperatur kleiner als die Zieltemperatur-Hysterese ist.
      Ist dies der Fall setzen wir den heizungCode auf HEIZUNG_AUS damit nachfolgender Code
      seine "reguläre Arbeit" durchführen kann (Heizung in intervallen ein/ausschalten)
      */
      if(heizungCode == HEIZUNG_WARTET) {
        if(sensordaten.temperatur < (hbsettings.zieltemperatur-hbsettings.hysterese)) {
            heizungCode = HEIZUNG_AUS;
            heizungSteuerung();
            #ifdef HBDEBUG
            Serial.println(F("Hysterese unterschritten"));
            #endif
        }
      } 
   
      else
      /*
      HEIZUNG:
      Wenn obriges nicht zutraf, prüfen wir, ob die Heizung AUS ist.
      Ist sie dies und ist die "Pausendauer" abgelaufen, schalten wir
      die Heizung an.
      */  
      if(heizungCode != HEIZUNG_AN) {
        if((heizungIntervall + (hbsettings.heizungPausedauer*1000)) < millis()) {
              heizungCode = HEIZUNG_AN;
              heizungSteuerung();
              heizungIntervall=millis();
              #ifdef HBDEBUG
              Serial.println(F("Heizung an"));
              #endif
        }
      }
    
      /*
      HEIZUNG:
      Ist die Heizung AN prüfen wir, ob die "Laufdauer" abgelaufen ist und schalten in die Konvektionspause.
      */
      else {
        if((heizungIntervall + (hbsettings.heizungLaufdauer*1000)) < millis()) {
          heizungCode = HEIZUNG_KONVEKTION;
          heizungSteuerung();
          heizungIntervall=millis();
          #ifdef HBDEBUG
          Serial.println(F("Heizung aus"));
          #endif
        }
      };
        

    /*
    VENTILATOR:
    Wenn obriges nicht zutraf, prüfen wir, ob die Ventilator AUS ist.
    Ist sie dies und ist die "Pausendauer" abgelaufen, schalten wir
    den Ventilator an.
    */  
    if(ventilatorCode == VENTILATOR_AUS) {
      if(ventilatorIntervall + (hbsettings.ventilatorPausedauer*1000) < millis()){
            if(
              (state==STATE_COUNTDOWN && countdownMitHeizen && heizungCode == HEIZUNG_AN) 
              ||
              (state != STATE_COUNTDOWN)
            ) {
              ventilatorCode = VENTILATOR_AN;
              ventilatorSteuerung();
              ventilatorIntervall=millis();
              #ifdef HBDEBUG
              Serial.println(F("Ventilator an"));
              #endif
            }
       }
     }
    
     /*
     VENTILATOR:
     Ist der Ventilator AN prüfen wir, ob die "Laufdauer" abgelaufen ist und schalten ihn aus.
     */
     else {
       if(ventilatorIntervall + (hbsettings.ventilatorLaufdauer*1000) < millis()) {
         ventilatorCode = VENTILATOR_AUS;
         ventilatorSteuerung();
         ventilatorIntervall=millis();
         #ifdef HBDEBUG
         Serial.println(F("Ventilator aus"));
         #endif
       }
     }
  // } //State != CountDown     --- wieder entfernt weil jetzt im CountDown auch eine Warmhaltefunktion existiert

  /* Da wir die Heizung und die Lüftersteuerung der Spannung nachführen müssen wir
  * regelmäßig die Subroutinen aufrufen, damit die PWM-Werte aktualisiert werden.
  */
  
  if((uint8_t)sensordaten.spannung!=sensordaten.altspannung) {
    heizungSteuerung();
    ventilatorSteuerung();
    sensordaten.altspannung=sensordaten.spannung;
  }     
 } // if (fehlerCode =0)
 
 
 /* Beim letzten Wechsel der Hintergrundbeleuchtung wurde eine Zeit gesetzt, ab wann
  * der nächste Wechsel stattfinden soll. Ist diese Zeit erreicht oder Überschritten
  * wird der nächste Wechsel durchfgeführt.
  */
 if(LEDZielZeit <= millis()) { LEDSteuerung(); }
 
 
 /* Das Pre-Menü soll schon Erscheinen, wenn lange genug der EncoderButton gedrückt gehalten wurde
  * ohne dass erst drauf gewartet wird, dass der Button wieder losgelassen wurde.
  * Hierzu werden wir aus, ob der Button gedrückt gehalten wurde und ob die nötige Zeit verstrichen ist
  * und rufen dann das PRe-Menü auf.
  */
  if(EncoderButtonPressed && (encoderPressed+EncoderPresstoEnterMenu < millis()) ) {
       if(state==STATE_NORMAL) { 
         
         erstelleTemporaereSettings();
         currentItem = &premenu1;
         state = STATE_MENU;
         lcd.clear();
         zeigeMenu();
       }
       if(state==STATE_COUNTDOWN) {
         deaktiviereCountdown();
         lcd.clear();
       };
       EncoderButtonPressed=false;
  }
  
}




/**--------------------------------------------------------------------------------------------------------------------------------
 * encoderCallback ist die CallBack-Funktion 
 * für die Encoder-Bibliothek. Passiert am Encoder
 * etwas, wird diese Methode aufgerufen.
 * @param byte b : beinhaltet die Aktion die am Encoder aufgetreten ist
 */
void encoderCallback(byte b) {
  encoderSignal=b;
  
}


/**--------------------------------------------------------------------------------------------------------------------------------
 * handleEncoderFlag - wertet den letzten Encoder-Zustand aus, der via encoderCallback() geliefert wurde.
 * Die Trennung zwischen ISR-naher Methode und dieser ist nötig, um die ISRs so kurz wie möglich zu halten.
 */
void handleEncoderFlag() {
  LEDSteuerung(LED_AN);
 
  if(encoderSignal==2 || encoderSignal == 1){       //Drehen rechts oder Links
    switch(state) {
     case STATE_MENU:currentItem=(encoderSignal==2?(*currentItem).next:(*currentItem).prev); lcd.clear(); zeigeMenu(); hbtone->play(0); break; 
     case STATE_MENU_CHANGEVALUE: changeSettingsValue(encoderSignal);  zeigeMenuWerte(true); hbtone->play(0); break;
     #ifndef D4ZEILEN
     case STATE_NORMAL:if(displayPage==DISPLAY_MAIN)displayPage=DISPLAY_SECONDARY;else displayPage=DISPLAY_MAIN; break;
     #endif  
  }
  
  }
  else if(encoderSignal==3) {    //Drücken
    //switch(state) {
    //  case STATE_NORMAL: if(!EncoderButtonPressed) {EncoderButtonPressed=true; encoderPressed=millis();} break;
    //}
    if(!EncoderButtonPressed) {EncoderButtonPressed=true; encoderPressed=millis();}
    
  } else if(encoderSignal==4) { //Loslassen
      if(EncoderButtonPressed == false) {
        encoderSignal=0;
        return;      //Wurde der Button vorher gar nicht gedrückt ODER wurde das Drücken an anderer Stelle "gelöscht" machen wir hier nichts weiter
      }
      
      //if(EncoderButtonPressed == true) {
        EncoderButtonPressed=false;
       // return;      //Wurde der Button vorher gar nicht gedrückt ODER wurde das Drücken an anderer Stelle "gelöscht" machen wir hier nichts weiter
      //}
      switch(state) {
          case STATE_NORMAL: if(fehlerUnbestaetigt) fehlerUnbestaetigt=false;/*vorhandene Fehler quittieren*/ 
                             if(warnschwelleUnbestaetigt) warnschwelleUnbestaetigt=false;
                             break;

  /*      case STATE_COUNTDOWN: if(encoderPressed+EncoderPresstoExitCountdown < millis()){
                                    deaktiviereCountdown();
                                    lcd.clear();
                                    
                         }
                         break;
                         
*/
        case STATE_MENU: if((*currentItem).subMenu!=NULL) { 
                              if((*currentItem).subMenu==&defaultscreen){  
                                      state=STATE_NORMAL; 
                                      lcd.clear(); 
                              }else {
                                currentItem = (*currentItem).subMenu;
                                lcd.clear();
                                zeigeMenu();
                                if( (*currentItem).next==currentItem && (*currentItem).prev==currentItem ) {    //Hat das neue Submenu-Item weder einen Vorgänger noc heinen Nachfolger (Einzigstes Element) dann soll sofort der Wert verändert werden können
                                  state=STATE_MENU_CHANGEVALUE; 
                                  zeigeMenuWerte(true);
                                }
                                
                              }
                        } else if((*currentItem).value!=NULL && (*currentItem).type!=20)  {    //Wert ist nicht NULL und Typ ist nicht 20 (Info-Eintrag)
                           state=STATE_MENU_CHANGEVALUE; 
                           zeigeMenuWerte(true);
                           
                        } /*else if ((*currentItem).next==currentItem && (*currentItem).prev==currentItem) {
                           state=STATE_MENU_CHANGEVALUE; 
                           zeigeMenuWerte(true);
                        }*/
                        break;
 
 
        case STATE_MENU_CHANGEVALUE:
                         state=STATE_MENU; 
                         if(setzeStandartWerte)    {  ladeStandardwerte();  erstelleTemporaereSettings();  setzeStandartWerte=false; bestaetigeAktion(STATE_INVALID /*in aktueller Anzeige bleiben*/); }
                         if(setzeVorherigeWerte)   {  erstelleTemporaereSettings();    setzeVorherigeWerte=false;                    bestaetigeAktion(STATE_INVALID /*in aktueller Anzeige bleiben*/); }
                         if(speichereAktuelleWerte){  uebernehmeTemporaereSettings();  speichereAktuelleWerte=false;                 bestaetigeAktion(STATE_NORMAL); }
                         if(verbrauchZuruecksetzen){  akkuZuruecksetzen(); verbrauchZuruecksetzen=false;  currentItem = &premenu1;  zeigeMenu(); bestaetigeAktion(STATE_INVALID /*in aktueller Anzeige bleiben*/); }
                         else if(currentItem == &bestaetigeReset){ state=STATE_MENU; currentItem = &premenu1; zeigeMenu(); }
                        
                         if(starteCountdown){   
                           if((countdownZeitStunden+countdownZeitMinuten)>0) {
                             lcd.clear();
                             state = STATE_COUNTDOWN;
                             initialisiereCountDown();      //CountDown beginnen, Heizung und Lüfter aus, etc
                           } else {
                             state=STATE_MENU;  
                           }
                           starteCountdown = false;
                         
                         }
                         
                         if(currentItem == &timerEinstellen4) {
                           currentItem = &premenu2;
                           zeigeMenu();
                         }
                         
                         zeigeMenuWerte(false);
                         break;
    }
    hbtone->play(1);

  }
  
  //Encoder-Flag zurück setzen
  encoderSignal=0;
}


#ifdef D4ZEILEN
/**--------------------------------------------------------------------------------------------------------------------------------
 * Zeigt den Haubtbildschirm an inklusive aller Informationen
 *
 */
void zeigeHauptschirm() {
  /*
#ifndef D4ZEILEN
  if(displayPage!=vorherigeDisplayPage) {
    lcd.clear();
    vorherigeDisplayPage = displayPage;
  }
#endif
*/
  lcd.setCursor(0,0);
  lcd.write((byte)1);
  lcd.print(padspace(sensordaten.temperatur,1,4));
  lcd.write((byte)0);
  lcd.setCursor(10,0);
  lcd.write((byte)2);
  lcd.print(padspace(hbsettings.zieltemperatur,0,2));
  lcd.write((byte)0);


  lcd.setCursor(0,1);
  lcd.write((byte)4);
  lcd.print(padspace(sensordaten.spannung,1,4));
  lcd.print(F("V "));

  

  
  if(!externeVersorgung) {
    lcd.setCursor(8,1);
    lcd.print(padspace(sensordaten.verbrauch,0,5));
    lcd.print(F("mAh"));
  
  
    lcd.setCursor(0,2);
    lcd.write((byte)3);
    for(uint8_t i=1; i<16; i++) {
      if(sensordaten.prozent>=7*i) lcd.print(char(255)); else lcd.print(" ");
    }
  } else {
   lcd.setCursor(0,2);
   lcd.print((char)0xff);
   lcd.print(externeQuelle);
   lcd.print((char)0xff);
  }


  lcd.setCursor(0,3);
    if(fehlerCode!=ERROR_NOERROR){
       if(bildschirmFehlerWechsel<millis()) zeigeFehlerTextMeldung();
    } else if(warnschwelleErreicht && warnschwelleUnbestaetigt) {
      lcd.write((byte)5);
      lcd.print(F(" Akku "));
      lcd.write((byte)5);
      lcd.print(leer16Zeichen);
    } else {
      switch(heizungCode){
        case HEIZUNG_AUS:          lcd.print(F("    inaktiv"));    break;
        case HEIZUNG_AN:           lcd.print(F("     heizen"));    break;
        case HEIZUNG_KONVEKTION:   lcd.print(F("   konvektion"));  break;
        case HEIZUNG_WARTET:       lcd.print(F("     warten"));    break;
      }
      lcd.print(leer16Zeichen);
    }
    
  }

#else

/**--------------------------------------------------------------------------------------------------------------------------------
 * Zeigt den Haubtbildschirm an inklusive aller Informationen
 *
 */
void zeigeHauptschirm() {
  if(displayPage!=vorherigeDisplayPage) {
    lcd.clear();
    vorherigeDisplayPage = displayPage;
  }
  
  if(displayPage==DISPLAY_MAIN) {

    lcd.setCursor(0,0);
    lcd.write((byte)1);
    lcd.print(padspace(sensordaten.temperatur,1,4));
    lcd.write((byte)0);
    
    lcd.setCursor(9,0);
    lcd.write((byte)2);
    lcd.print(padspace(hbsettings.zieltemperatur,0,2));
    lcd.write((byte)0);
  
   
    lcd.setCursor(0,1);
    lcd.write((byte)4);
    lcd.print(padspace(sensordaten.spannung,1,4));
    lcd.print(F("V "));
    
    if(!externeVersorgung) {
      lcd.setCursor(9,1);
      lcd.write((byte)3);
      lcd.print(padspace(sensordaten.prozent,0,3));
      lcd.print(F("%"));
    } else {
      lcd.setCursor(9,1);
      lcd.print(F("Extern"));
    }  
   
    lcd.setCursor(15,0);
    lcd.write((byte)6);
    lcd.setCursor(15,1);
    lcd.write((byte)6);
    
  } else {
    lcd.setCursor(0,0);
    lcd.write((byte)7);
    lcd.setCursor(0,1);
    lcd.write((byte)7);
    
    if(!externeVersorgung) {
      lcd.setCursor(2,0);
      lcd.print(padspace(sensordaten.verbrauch,0,0));
      lcd.print(F("mAh"));
    }
    
    lcd.setCursor(2,1);
    if(fehlerCode!=ERROR_NOERROR){
       if(bildschirmFehlerWechsel<millis()) zeigeFehlerTextMeldung();
    } else if(warnschwelleErreicht && warnschwelleUnbestaetigt) {
      lcd.write((byte)5);
      lcd.print(F(" Akku "));
      lcd.write((byte)5);
      lcd.print(leer16Zeichen);
    } else {
      switch(heizungCode){
        case HEIZUNG_AUS:          lcd.print(F("inaktiv"));   break;
        case HEIZUNG_AN:           lcd.print(F("heizen"));    break;
        case HEIZUNG_KONVEKTION:   lcd.print(F("konvektion"));   break;
        case HEIZUNG_WARTET:       lcd.print(F("warten"));    break;
      }
      lcd.print(leer16Zeichen);
    }
    
  }
  
  

}

#endif



/**--------------------------------------------------------------------------------------------------------------------------------
 * initialisiereCountDown - schaltet Heizung und Lüfter ab
 * und bereitet den Countdown vor.
 */
void initialisiereCountDown(){
    state=STATE_COUNTDOWN;
    if(countdownMitHeizen) {
      heizungCode = HEIZUNG_AN;
      ventilatorCode = VENTILATOR_AN;
    } else {
      heizungCode = HEIZUNG_AUS;
      ventilatorCode = VENTILATOR_AUS;
    }
  
    heizungSteuerung();
    ventilatorSteuerung();
    countdownStart=millis();
    lcd.clear(); 
}


/**--------------------------------------------------------------------------------------------------------------------------------
 * deaktiviereCountdown - schaltet den CountDown ab
 * und setzt inige Werte zurück
 */
void deaktiviereCountdown(){
  starteCountdown=false;
  countdownZeitStunden = 0;
  countdownZeitMinuten = 0;
  countdownZeitSekunden = 0;
  state = STATE_NORMAL;
   
}


/**--------------------------------------------------------------------------------------------------------------------------------
 * zeigeMenu stellt den aktuellen Menüeintrag dar und ruft 
 * eine Hilfsfunktion auf, um den dazugehörigen Wert dar zu stellen.
 */
void zeigeMenu(){
  lcd.setCursor(0,0);
  lcd.print((*currentItem).text);
  lcd.print(leer16Zeichen);
  zeigeMenuWerte(false);
}





/**--------------------------------------------------------------------------------------------------------------------------------
 * zeigeMenuWerte stellt die zum Menüeintrag passenden Werte dar
 * @param boolean active  =  wenn true wird ein Pfeil angezeigt um zu Symbolieiseren, dass Wert geändert werden kann
 */
void zeigeMenuWerte(boolean active) {
  if(zeigeBestaetigung)return;        //Wird gerade noch eine Bestätigung angezeigt, mache hier gar nichts
  //die untere Zeile enthält sowohl das Zeichen ob Auswahl aktiv ist ("Enter-Pfeil") als auch den Wert.
  //um keine Reste des alten Wert oder das Auswahl-Aktiv zeichen zu sehen obwohl es weg sein sollte löschen wir die untere Zeile
  lcd.setCursor(0,1);                
  lcd.print(leer16Zeichen);           
  if((*currentItem).value!=NULL) {
   if(active) {    //Wurde der aktuell angezeigte Wert zum Ändern gewählt? Dann zeige den Enter-Pfeil.
     lcd.setCursor(0,1);
     lcd.print(char(126));
   }
   lcd.setCursor(3,1);
   switch((*currentItem).type) {
     case 0:  lcd.print( *(uint8_t*)(*currentItem).value ); lcd.write((byte)0); break;            //Typ 0 = uint8_t Temperatur
     case 1:  lcd.print( *(unsigned int*)(*currentItem).value ); lcd.print(F("mAh")); break;      //Typ 1 = unsigned Long Milliampere
     case 2:  lcd.print( *(boolean*)(*currentItem).value?F("Ja"):F("Nein") ); break;              //Typ 2 = boolean "Ja" oder "Nein"
     case 4:  lcd.print( *(uint8_t*)(*currentItem).value ); lcd.print(F("sek")); break;           //Typ 4 = uint8_t Zeit in Sekunden
     case 5:  lcd.print( *(uint8_t*)(*currentItem).value ); break;                                //Typ 5 = uint8_t Zahl ohne Einheit
     case 6:  lcd.print( akku_names[  hbtempsettings.batterietyp ] ); break;                      //Typ 6 = uint8_t Index auf Batterietyp-Array
     case 7:  lcd.print( akku_zellen[ hbtempsettings.batterietyp ][ hbtempsettings.zellenzahl ]); break;                    //Typ 7 = uint8_t Index auf Zellenzahl-Array
     case 8:  lcd.print("["); lcd.print(countdownZeitStunden); lcd.print(F("h]: ")); lcd.print(countdownZeitMinuten ); lcd.print(F("min")); break;
     case 9:  lcd.print(" "); lcd.print(countdownZeitStunden); lcd.print(F("h :[")); lcd.print(countdownZeitMinuten ); lcd.print(F("min]")); break;                        //Typ 8 = uint8_t Zeit in minuten und Stunden
     case 10: 
     case 11: lcd.print( *(uint8_t*)(*currentItem).value ); break;            //Typ 10 = uint8_t DisplayHelligkeit, Type 11 = Kalibrierwert
     case 20: lcd.setCursor(0,1); lcd.print( (char*)(*currentItem).value ); break;                                       //Typ 20 = unveränderlicher String, nur Info
     //case 21: lcd.setCursor(0,1); lcd.print( *(uint8_t*)(*currentItem).value ); break;                                       //Typ 21 = unveränderliche Zahl, nur Info 
   }
 } 
}



/**--------------------------------------------------------------------------------------------------------------------------------
 * Zeigt eine Bestätigung im Display an, setzt <i>zeigeBestaetigung</i> aud TRUE
 * und merkt sich, wann die Bestätigung angezeigt wurde. Die loop-Funktion soll
 * dann nach Ablauf einer gegebenen Zeit die Standart-Anzeige wiederherstellen.
 */
void bestaetigeAktion(uint8_t newState){
  state_afterConfirm = newState;
  lcd.setCursor(0,1);
  lcd.print(F("     - OK -     "));    //Zeile löschen
  zeigeBestaetigung=true;
  bestaetigungsAnzeige=millis();
  
}





/**--------------------------------------------------------------------------------------------------------------------------------
 * ladeEEPROMWerte - läd die Settings aus dem EEPROM.
 * Sollte festgestellt werden, dass die Settings aus dem EEPROM leer sind (erster Start)
 * werden die Default-Werte geladen und in die Settings gespeichert.
 *
 *@return boolean : TRUE wenn Standartwerte wiederhergestellt wurden, FALSE wenn nicht
 */
boolean ladeEEPROMWerte(){
  boolean ret = false;
  eeprom->readSettings(&hbsettings);
  
  /*Prüfen, ob die Settings "jungfräulich" sind. 
    Dies ist der Fall wenn noch nie eine Settings in den EEPROM gespeichert 
    wurde (beim Ersten Start ist dies der Fall). Denn dann stehen alle Werte auf "0"
  */
  if(hbsettings.virginity == false){
    ladeStandardwerte();
    hbsettings.virginity = true;
    eeprom->writeSettings(&hbsettings);
    #ifdef HBDEBUG
      Serial.println(F("Erstinitialisierung"));
    #endif  
  }
  
  if(hbsettings.version != settings_version){
    eeprom->clearRom();
    ladeStandardwerte();
    hbsettings.version = settings_version;
    eeprom->writeSettings(&hbsettings);
    ret=true;
    #ifdef HBDEBUG
      Serial.println(F("Settings-Update")); 
    #endif  
  }

  unterSpannungsGrenze=akku_spannungen[hbsettings.batterietyp][hbsettings.zellenzahl];
  
  return ret;
}






/**--------------------------------------------------------------------------------------------------------------------------------
 *  ladeStandardwerte - Läd die Default-Werte in die Settings-Struktur
 */
void ladeStandardwerte() {
  hbsettings.zieltemperatur       = vorgabe_zieltemperatur;
  hbsettings.hysterese            = vorgabe_hysterese;
  hbsettings.batterietyp          = vorgabe_batterietyp;             //Index auf Array
  hbsettings.zellenzahl           = vorgabe_zellenzahl;              //Index auf Array
  hbsettings.kapazitaet           = vorgabe_kapazitaet;
  hbsettings.abschaltgrenze       = vorgabe_abschaltgrenze;
  hbsettings.sollWarnen           = vorgabe_sollWarnen;
  hbsettings.warnschwelle         = vorgabe_warnschwelle;
  hbsettings.sollSpeichern        = vorgabe_sollSpeichern;            //Speichern in EEPROM
  hbsettings.heizungLaufdauer     = vorgabe_heizungLaufdauer;
  hbsettings.heizungPausedauer    = vorgabe_heizungPausedauer;
  hbsettings.pwmboost             = vorgabe_pwmboost;
  hbsettings.ventilatorLaufdauer  = vorgabe_ventilatorLaufdauer;
  hbsettings.ventilatorPausedauer = vorgabe_ventilatorPausedauer;
  hbsettings.alarmTon             = vorgabe_alarmTon;
  hbsettings.displayHelligkeit    = vorgabe_displayHelligkeit;
  hbsettings.kalibrierWert        = vorgabe_kalibrierWert;
  hbsettings.displayDauerAn       = vorgabe_displayDauerAn;
  hbsettings.warmhalteTemperatur  = vorgabe_warmhalteTemperatur;
}






/**--------------------------------------------------------------------------------------------------------------------------------
 * Kopiert die aktuellen Settings in eine
 * temporäre Kopie (zum Einstellen im Setup)
 */
void erstelleTemporaereSettings() {
  hbtempsettings = hbsettings;
}






/**--------------------------------------------------------------------------------------------------------------------------------
 * Kopiert die temporären Settings in
 * die originale Struktur
 */
void uebernehmeTemporaereSettings() {
  hbsettings = hbtempsettings;
  unterSpannungsGrenze=akku_spannungen[hbsettings.batterietyp][hbsettings.zellenzahl];
  eeprom->writeSettings(&hbsettings);
  batterie->setKapazitaet(hbsettings.kapazitaet);
  batterie->setKalibrierwert(hbsettings.kalibrierWert);
  heizungSteuerung();    //Falls man was am PWMBoost geändert hat -> Werte direkt übernehmen.
}





/**--------------------------------------------------------------------------------------------------------------------------------
 * Läd die Icons aus PROGMEM in den RAM des Displays
 */
void ladeIcons() {
  lcd.createChar(0, gradcelsiusIcon);
  lcd.createChar(1, thermometerIcon);
  lcd.createChar(2, zieltemperaturIcon);
  lcd.createChar(3, batterieIcon);
  lcd.createChar(4, spannungIcon);
  lcd.createChar(5, batterieLeerIcon);
  lcd.createChar(6, rightIcon);
  lcd.createChar(7, leftIcon);
}





/**--------------------------------------------------------------------------------------------------------------------------------
 * changeSettingsValue - in Abhängigkeit von aktuellen Menüpunkt ändert diese Funktion den dazugehörigen Wert
 * @ param byte b : beschreibt die Drehrichtung des Encoders
 */
void changeSettingsValue(byte b) {    //B0 = RECHTS
   boolean isFast = false;
   if((millis()-letzterDreh)<200) {
     isFast=true;
   };
   letzterDreh=millis();  
   switch((*currentItem).type) {
     case 11: batterie->setKalibrierwert(hbtempsettings.kalibrierWert);      //Kalibriermenü
     case 4:                                //Typ 4 = uint8_t Zeit in Sekunden  
     case 5:                                //Typ 5 = uint8_t Zahl ohne Einheit

     case 0:  if(b==2) {                                //Typ 0 = uint8_t Temperatur
                 if( (*(uint8_t*)(*currentItem).value) < (*currentItem).maxWert) (*(uint8_t*)(*currentItem).value)+=isFast?schrittweite_temp_max:schrittweite_temp_min; 
              } else {
                 if( (*(uint8_t*)(*currentItem).value) > (*currentItem).minWert) (*(uint8_t*)(*currentItem).value)-=isFast?schrittweite_temp_max:schrittweite_temp_min;
              }
              if( (*(uint8_t*)(*currentItem).value) > 200) *(uint8_t*)(*currentItem).value = (*currentItem).minWert;  //Beim Unterschreiten (Erg. kleiner 0) wird der Wert als großer positiver Wert interpretiert. Soll aber Null serin.
              if( (*(uint8_t*)(*currentItem).value) > (*currentItem).maxWert) *(uint8_t*)(*currentItem).value = (*currentItem).maxWert;
              if( (*(uint8_t*)(*currentItem).value) < (*currentItem).minWert) *(uint8_t*)(*currentItem).value = (*currentItem).minWert;
              break;
     
    
     case 1:  if(b==2) {                                //Typ 1 = unsigned Long Milliampere
                 if( *(unsigned int*)(*currentItem).value < (*currentItem).maxWert) (*(unsigned int*)(*currentItem).value)+=isFast?schrittweite_mah_max:schrittweite_mah_min; 
              } else {
                 if( *(unsigned int*)(*currentItem).value > (*currentItem).minWert) (*(unsigned int*)(*currentItem).value)-=isFast?schrittweite_mah_max:schrittweite_mah_min;
              }
              if( (*(unsigned int*)(*currentItem).value) > (*currentItem).maxWert + 500) *(unsigned int*)(*currentItem).value = (*currentItem).minWert;  //Beim Unterschreiten (Erg. kleiner 0) wird der Wert als großer positiver Wert interpretiert. Soll aber Null serin.
              if( (*(unsigned int*)(*currentItem).value) > (*currentItem).maxWert) *(unsigned int*)(*currentItem).value = (*currentItem).maxWert;
              if( (*(unsigned int*)(*currentItem).value) < (*currentItem).minWert) *(unsigned int*)(*currentItem).value = (*currentItem).minWert;

              if(hbtempsettings.abschaltgrenze>hbtempsettings.kapazitaet)hbtempsettings.abschaltgrenze=hbtempsettings.kapazitaet;            //Aktuellen Wert der Abschaltgrenze anpassen wenn er über der Kapazität liegt
              if(hbtempsettings.warnschwelle>hbtempsettings.abschaltgrenze-10)hbtempsettings.warnschwelle=hbtempsettings.abschaltgrenze-10;  //Aktuellen Wert der Warnschwelle anpassen wenn er über der Abschaltgrenze - 10 liegt

              if(currentItem==&batt3){ //Ändern der Kapazität
                *(unsigned int*)batt4.maxWert=hbtempsettings.kapazitaet;                         //Obergrenze Abschaltschwelle anpassen
                *(unsigned int*)batt6.maxWert=hbtempsettings.abschaltgrenze-10;                  //Obergrenze Warnschwelle anpassen
              };
              
              if(currentItem==&batt4){ //Ändern der Abschaltschwelle
                *(unsigned int*)batt6.maxWert=hbtempsettings.abschaltgrenze-10;                  //Obergrenze Warnschwelle anpassen
              };
              break;  
     
                               
     case 2:  *(boolean*)(*currentItem).value = !*(boolean*)(*currentItem).value; break;                              //Typ 2 = boolean "Ja" oder "Nein"

 
     case 6:  if(b==2) {                                //Typ 6 = uint8_t Index auf Batterietyp-Array
                 if( (*(uint8_t*)(*currentItem).value) < (*currentItem).maxWert) (*(uint8_t*)(*currentItem).value)++;
              } else {
                 if( (*(uint8_t*)(*currentItem).value) > (*currentItem).minWert) (*(uint8_t*)(*currentItem).value)--;
              }
              hbtempsettings.zellenzahl=0;      //Wenn eun neuer Akku gewählt wird setzen wir das Zellenzahl-Index auf 0 um kein Out-Of-Bounds zu erzeugen
              if( (*(uint8_t*)(*currentItem).value) > (*currentItem).maxWert) *(uint8_t*)(*currentItem).value = (*currentItem).maxWert;
              if( (*(uint8_t*)(*currentItem).value) < (*currentItem).minWert) *(uint8_t*)(*currentItem).value = (*currentItem).minWert;
              break;

              
     case 7:  if(b==2) {                                //Typ 7 = uint8_t Index auf Zellenzahl-Array
                 //ToDo : Überarbeiten
                 uint8_t tempmax=0;
                 switch(hbtempsettings.batterietyp) {
                   case 0:  tempmax= anzahlLipoZellen;  break;
                   case 1:  tempmax= anzahlLifeZellen;  break;
                   case 2:  tempmax= anzahlLiIonZellen; break;
                   case 3:  tempmax= anzahlNixxZellen;  break;
                   case 4:  tempmax= anzahlBleiZellen;  break;
                 }
                 
                 if( (*(uint8_t*)(*currentItem).value) < tempmax-1) (*(uint8_t*)(*currentItem).value)++; 
              } else {
                 if( (*(uint8_t*)(*currentItem).value) > 0) (*(uint8_t*)(*currentItem).value)--;
              }
              if( (*(uint8_t*)(*currentItem).value) > (*currentItem).maxWert) *(uint8_t*)(*currentItem).value = (*currentItem).maxWert;
              if( (*(uint8_t*)(*currentItem).value) < (*currentItem).minWert) *(uint8_t*)(*currentItem).value = (*currentItem).minWert;
              break;
             
              
    
    
      case 8:  if(b==2) {                                //Typ 8 = unsigned int timer Countdown Stunden
                 if( (*(uint8_t*)(*currentItem).value < (*currentItem).maxWert)) (*(uint8_t*)(*currentItem).value)+=isFast?3:1;
              } else {
                 if( (*(uint8_t*)(*currentItem).value > (*currentItem).minWert)) (*(uint8_t*)(*currentItem).value)-=isFast?3:1;
              }
              if( (*(uint8_t*)(*currentItem).value) > 100) *(uint8_t*)(*currentItem).value = (*currentItem).minWert;  //Beim Unterschreiten (Erg. kleiner 0) wird der Wert als großer positiver Wert interpretiert. Soll aber Null serin.
              if( (*(uint8_t*)(*currentItem).value) > (*currentItem).maxWert) *(uint8_t*)(*currentItem).value = (*currentItem).maxWert;
              if( (*(uint8_t*)(*currentItem).value) < (*currentItem).minWert) *(uint8_t*)(*currentItem).value = (*currentItem).minWert;
              break;
              
              
     case 9:  if(b==2) {                                //Typ 9 = unsigned int timer Countdown Minuten
                 if( (*(uint8_t*)(*currentItem).value < (*currentItem).maxWert)) (*(uint8_t*)(*currentItem).value)+=isFast?5:1;
              } else {
                 if( (*(uint8_t*)(*currentItem).value > (*currentItem).minWert)) (*(uint8_t*)(*currentItem).value)-=isFast?5:1;
              }
              if( (*(uint8_t*)(*currentItem).value) > 100) *(uint8_t*)(*currentItem).value = (*currentItem).minWert;  //Beim Unterschreiten (Erg. kleiner 0) wird der Wert als großer positiver Wert interpretiert. Soll aber Null serin.
              if( (*(uint8_t*)(*currentItem).value) > (*currentItem).maxWert) *(uint8_t*)(*currentItem).value = (*currentItem).maxWert;
              if( (*(uint8_t*)(*currentItem).value) < (*currentItem).minWert) *(uint8_t*)(*currentItem).value = (*currentItem).minWert;
              break;
              
     case 10:  if(b==2) {                                //Typ 10 = Helligkeit der Displaybeleuchtung
                 if( (*(uint8_t*)(*currentItem).value) < (*currentItem).maxWert) (*(uint8_t*)(*currentItem).value)+=1; 
              } else {
                 if( (*(uint8_t*)(*currentItem).value) > (*currentItem).minWert) (*(uint8_t*)(*currentItem).value)-=1;
              }
              if( (*(uint8_t*)(*currentItem).value) > (*currentItem).maxWert) *(uint8_t*)(*currentItem).value = (*currentItem).maxWert;
              if( (*(uint8_t*)(*currentItem).value) < (*currentItem).minWert) *(uint8_t*)(*currentItem).value = (*currentItem).minWert;
              
              SoftPWMSetValue(BACKGROUNDPWM, 105+(15*hbtempsettings.displayHelligkeit) );    //Die temporäre Helligkeit direkt anwenden.
              break;          
   }
}






/**--------------------------------------------------------------------------------------------------------------------------------
 * padspace - füllt eine übergebene Zahl
 * mit Leerzeichen auf bis sie der gegebenen Länge entspricht
 * und gibt das als String zurück. Floatwerte werden auf die
 * angegebene Nachkommastellenzahl gerundet.
 * @param float val: Zahlenwert der aufgefüllt werden soll
 * @param uint8_t nachkomma: Anzahl der darzustellenden Nachkommastellen
 * @param uint8_t len: Länge auf die aufgefällt werden soll
 * @return String: auf die Länge aufgefüllter String
 */
String padspace(float val, uint8_t nachkomma, uint8_t len) {
 //char charBuf [len+1];   <- eventuell doof wenn Wert länger ist als len+1   !!
 char charBuf [10];
 dtostrf(val, len, nachkomma, charBuf);
 return charBuf;
}

/**--------------------------------------------------------------------------------------------------------------------------------
 *
 */
 /*
String padzero(int val, uint8_t len) {
  String ret = ""+val;
  while(ret.length()<len) ret="0"+ret;
  return ret;
}
*/



/**--------------------------------------------------------------------------------------------------------------------------------
 * akkuZuruecksetzen - setzt den verbrauch auf 0 und speichert den Wert in den EEPRom
 */
void akkuZuruecksetzen(){
  batterie->reset();
  sensordaten.verbrauch = 0;
  warnschwelleErreicht=false;
  warnschwellenMerker=false;
  if((fehlerCode & BATTERIE_ERROR_ABSCHALTSCHWELLE) > 0) fehlerCode ^= BATTERIE_ERROR_ABSCHALTSCHWELLE;   //Eventuellen Abschaltschwellenfehler auch resetten
  if(fehlerCode==ERROR_NOERROR) { fehlerUnbestaetigt = false;};
  if(hbsettings.sollSpeichern)eeprom->writeCapacity(&sensordaten.verbrauch);
}






/**--------------------------------------------------------------------------------------------------------------------------------
 * ladeLetztenVerbrauch - ließt den gespeicherten Verbrauch aus dem EEPROM ein
 * und übergibt den Wert an die Baterie-Bibliothek als "Startwert"
 */
void ladeLetztenVerbrauch(){
  eeprom->readCapacity(&sensordaten.verbrauch);
  batterie->setVerbrauch(sensordaten.verbrauch);
}




/**--------------------------------------------------------------------------------------------------------------------------------
 * zeigeCountdown - gibt die aktuelle verbleibende Zeit bis zum Ablauf des CountDowns
 * im Display an.
 */
void zeigeCountdown(){
  lcd.setCursor(0,0);
  lcd.print(F("Timer "));
  if(countdownZeitStunden<10)lcd.print("0");
  lcd.print(countdownZeitStunden);
  lcd.print(":");
  if(countdownZeitMinuten<10)lcd.print("0");
  lcd.print(countdownZeitMinuten);
  lcd.print(":");
  if(countdownZeitSekunden<10)lcd.print("0");
  lcd.print(countdownZeitSekunden);
  lcd.print("h");
  

  
  if(externeVersorgung){
    lcd.setCursor(0,1);
    lcd.print((char)0xff);
    lcd.print(externeQuelle);
    lcd.print((char)0xff);
  } else {
    lcd.setCursor(0,1);
    lcd.write((byte)4);
    lcd.print(padspace(sensordaten.spannung,1,4));
    lcd.print(F("V "));
   
    lcd.setCursor(9,1);
    lcd.write((byte)3);
    lcd.print(padspace(sensordaten.prozent,0,3));
    lcd.print(F("%"));
  }
}





/**--------------------------------------------------------------------------------------------------------------------------------
 * ErrorCallBack - CallBack-Funktion für Fehler die igendwo (auch in Bibliotheken)
 * auftreten können.
 *
 * @param void* quelle : Pointer auf Instanz der Bibliothek ("this")
 * @param uint8_t error: Fehlercode
 */
 
void ErrorCallBack(void* quelle, uint16_t error) {
  //fehlerMeldung=error;
  if((error & fehlerCode)!=0 ) return;    //Fehler ist schon bekannt.
  if(error == BATTERIE_ERROR_WARNSCHWELLE) {
      if(warnschwelleErreicht)return;      //Wurde die Warnschwelle bereits erreicht? Dann brauchen wir nichts mehr zu tun.
      warnschwelleErreicht=true;
    } else {
      fehlerCode |= error;
      if(state==STATE_COUNTDOWN) deaktiviereCountdown();    //taucht der Fehler z.B. während des CountDown auf springen wir auf die Standard-Anzeige um die Fehler auch anzuzeigen
   }
   
   fehlerUnbestaetigt=true;
  
  
  #ifdef HBDEBUG
  Serial.print(F("ECB:"));
  if(quelle == tsensor) Serial.print(F("TSsensor "));
  else if(quelle == batterie)  Serial.print(F("Batterie "));
  Serial.println(error); 
  #endif
 
    /*
    switch(error) {
     case BATTERIE_ERROR_WARNSCHWELLE:     warnschwelleErreicht=true;                      break;
     case BATTERIE_ERROR_ABSCHALTSCHWELLE: fehlerCode |= error;           break;
     case BATTERIE_ERROR_UNTERSPANNUNG:    fehlerCode |= ERROR_UNTERSPANNUNG;              break;
     case BATTERIE_ERROR_UEBERSPANNUNG:    fehlerCode |= BATTERIE_ERROR_UEBERSPANNUNG;     break;
    }    */

}





/**--------------------------------------------------------------------------------------------------------------------------------
 * zaehleFehler - zählt die gesetzten Fehler-Bits
 * @return uint8_t : Anzahl der Fehler
 */
uint8_t zaehleFehler(){
 uint16_t temp = fehlerCode;
 uint8_t anzahl = 0;
 while(temp>0) {
  if(temp & 0x01) anzahl++;
  temp>>=1;
 }
 return anzahl;
}


/**--------------------------------------------------------------------------------------------------------------------------------
 * heizungSteuerung - in Abhängigkeit des Status für die Heizung wird 
 * die Heizung ein (mit PWM) oder ausgeschaltet.
 *
 */
void heizungSteuerung(){
  if(heizungCode==HEIZUNG_AN) {
    uint8_t index=0;
    if(sensordaten.spannung>12) index = (uint8_t)sensordaten.spannung-12;
    #ifdef HBDEBUG
      Serial.print(F("heizungsSteuerung:  U="));
      Serial.print(sensordaten.spannung);
      Serial.print(F(" Idx="));
      Serial.println(index);
    #endif
    uint16_t temppwm = hbsettings.pwmboost + pgm_read_byte_near(&heizungPwmListe[index]);
    if (temppwm>255) temppwm=255;
    SoftPWMSetValue(HEIZUNGPWM, temppwm);
    
  } else {
   SoftPWMSetValue(HEIZUNGPWM,0);
  }
}



/**--------------------------------------------------------------------------------------------------------------------------------
 * ventilatorSteuerung - in Abhängigkeit des Status für den Ventilator wird 
 * der Ventilator ein (mit PWM) oder ausgeschaltet.
 *
 */
void ventilatorSteuerung(){
 #ifdef HBDEBUG
   Serial.print(F("Ventilatorsteuerung(): Code"));
   Serial.println(ventilatorCode);
 #endif
 int pwmwert = 0;          // damit notfalls auch Werte über 255 und kleiner 0 rein passen
 
 if(ventilatorCode==VENTILATOR_AN) {
  if(sensordaten.spannung>12)  pwmwert= berechnePWMVentilator(sensordaten.spannung);   //Bei Spannungen über 12V berechnen wir PWM für die jeweilige Spannung
  else                         pwmwert= berechnePWMVentilator(12);                     //Bei Spannungen kleiner oder gleich 12V berechnen wir PWM als wären es 12V
 }
 if(pwmwert > 255 || pwmwert < 0)pwmwert = 0;                                          //Möglicherweise war die Spannung kurzzeitig zu hoch - die Berechnung wäre dann unzuverlässig. machen wir den Ventilator lieber aus
 SoftPWMSetValue(VENTILATORPWM,pwmwert);
}


/**--------------------------------------------------------------------------------------------------------------------------------
 *
 *
 */
void LEDSteuerung(uint8_t neuerCode) {
 if(LEDCode == LED_BLINK) return;  //Blinken=Fehler, und wenn es einen fehler gibt wollen wir das Blinken nicht mehr überschreiben
 cli();
 LEDCode = neuerCode;
 LEDSteuerung();
 sei();
}


/**--------------------------------------------------------------------------------------------------------------------------------
 * LED-Steuerung - steuert die Hintergrundbeleuchtung des Displays in
 * Abhängigkeit des gesetzten Status
 */
void LEDSteuerung() {
  #ifdef HBDEBUG
  Serial.print(F("Schalte LED, LEDCode:"));
  Serial.println(LEDCode);
  #endif
  switch(LEDCode) {
    case LED_DELAYING: LEDCode=LED_AUS; //und weiter mit nächstem Case
    case LED_AUS:   if(hbsettings.displayDauerAn) {
                        SoftPWMSetValue(BACKGROUNDPWM,helligkeitDauer);
                    } else {
                        /*bitClear(PORTC, 2);*/   
                        SoftPWMSetValue(BACKGROUNDPWM,0);                      LEDZielZeit=0xFFFFFFFF; 
                    }
                    break;
    case LED_AN:    if(state==STATE_NORMAL) 
                      SoftPWMSetValue(BACKGROUNDPWM,105+(15*hbsettings.displayHelligkeit));
                    else
                      SoftPWMSetValue(BACKGROUNDPWM,helligkeitImMenu);  //In allen Menüs (da wo die LED längere Zeit an ist) muss die Helligkeit beschränkt werden wegen der mangelnden Auslegung des Spannungswandlers.
                 
                    LEDZielZeit=millis()+LED_AN_INTERVALL; 
                    LEDCode=LED_DELAYING;
                    break;
    case LED_BLINK: if(fehlerUnbestaetigt) {
                          if(SoftPWMGetValue(BACKGROUNDPWM)==0){SoftPWMSetValue(BACKGROUNDPWM,155+(10*hbsettings.displayHelligkeit));}else {SoftPWMSetValue(BACKGROUNDPWM,0);} LEDZielZeit=millis()+LED_BLINK_INTERVALL;
                    } else {
                        SoftPWMSetValue(BACKGROUNDPWM,105+(15*hbsettings.displayHelligkeit));
                    };
                    break;
    
    
  }
}

/**--------------------------------------------------------------------------------------------------------------------------------
 * Gibt nacheinander alle vorhandenen Fehler als Text aus
 *
 */
void zeigeFehlerTextMeldung( ){
   bildschirmFehlerWechsel = millis()+bildschirmFehlerZeigeAnzahl;
   if(fehlerAnzeige == 0) {
    fehlerAnzeige++;
    lcd.print(zaehleFehler());
    lcd.print(" ");
    lcd.print(wortFehler);
   } else if(fehlerAnzeige == 1) {
    fehlerAnzeige++;
    lcd.print(fehlerXvonY);
    lcd.print(F("/"));
    lcd.print(zaehleFehler());
    lcd.print(F(":"));
   } else {
     

     //Finde den nächsten Fehler
     while((fehlerCode & (1<<fehlerBitPosition)) == 0){
       fehlerBitPosition++;
       if(fehlerBitPosition>15 || fehlerXvonY>=zaehleFehler()) {fehlerBitPosition=0; fehlerXvonY=0;}
     }
     switch (fehlerBitPosition) {
      //case 0:  0 == Warnschwelle erreicht - wird woanders behandelt
      case 1:  lcd.print(F("Batterie leer"));      break;
      case 2:  lcd.print(F("Unter"));    lcd.print(wortSpannung);   break;
      case 3:  lcd.print(F("Ueber"));    lcd.print(wortSpannung);   break;
      case 4:  lcd.print(F("Strom"));    lcd.print(F("zu hoch"));   break;
      //5 frei (reserviert für weiteren Fehler in der Batterie-Biblithek)
      case 6:  lcd.print(wortTempSens);  lcd.print(F(" fehlt"));    break;
      case 7:  lcd.print(wortTempSens);  lcd.print(F(" defekt"));   break;
      case 8:  lcd.print(wortTempSens);  lcd.print(F(" unbek."));   break;
      case 9:  lcd.print(wortTempSens);  lcd.print(F(" entf."));    break;
      case 12: lcd.print(F("! FET DEFEKT !"));    break;
      //Frei bis 15
      default:lcd.print(F("unbek. ")); lcd.print(wortFehler);     break;
     }
     fehlerXvonY++;
     fehlerAnzeige = 0;
     fehlerBitPosition++;   
   }
  lcd.print(leer16Zeichen);
}
















/*  PC Kommunikationsschnittstelle
 *  Die Folgenden Komponenten dienen für die Kommunikation zwischen PC und HeatBox
 *
 *  Die Funktionalität kann mittels Schalter deaktiviert werden
 */


#ifdef ENABLEPCCOM

void printLog(const __FlashStringHelper* text){
  Serial.print("L:");
  Serial.print(text);
  Serial.print(";");
}


/**--------------------------------------------------------------------------------------------------------------------------------
 * Gibt anhand der datenNummer den entsprechenden Datensatz via Serieller
 * Schnittstelle aus.
 */
void printDaten(byte datenNummer){
  Serial.print("D:");    //Markiert dass es sich um einen Datensatz (keine Debugmeldung) handelt
  
  Serial.print(datenNummer,DEC);
  Serial.print(":");
  switch(datenNummer){
    case DATA_HBVERSION:                   Serial.print(hbVersionString); break;
    case DATA_ALL:                         dump_settings(); break;
    case DATA_VOLTAGE:                     Serial.print(sensordaten.spannung);
    case DATA_TEMPERATURE:                 Serial.print(sensordaten.temperatur);   break;
    case DATA_HEIZUNG_STATUS:              Serial.print(heizungCode);   break;
    default: Serial.print(0);
  }
  Serial.print("\n");
}


void dump_settings(){
  uint16_t i;
  Serial.print(sizeof(hbsettings),DEC);
  Serial.print("#");
  for (i = 0 ; i < sizeof(hbsettings) ; i++)
    Serial.print(((char *)&hbsettings)[i]);
}

/**--------------------------------------------------------------------------------------------------------------------------------
 * Even-Handler - reagiert, sobald Daten über die Serielle Schnitzstelle 
 * empfangen wurden.
 */
void serialEvent() {
  while (Serial.available()) {
    // get the new byte:
    char inChar = (char)Serial.read();
    // add it to the inputString:
    inputString += inChar;
    // if the incoming character is a newline, set a flag
    // so the main loop can do something about it:
    if (inChar == '\n') {
      stringComplete = true;
    }
  }
}


/**--------------------------------------------------------------------------------------------------------------------------------
 * Wurde eine festgestellt, dass die empfangenen Daten (inputString) vollständig sind, wird diese Methode aufgerufen
 * Sie entscheidet, wie mit den empfangenen Daten weiter gearbeitet werden soll.
 *
 */
void handleSerialInput(){
    if(inputString[0] == 'g'){  //get Data
        printDaten(inputString[1]);
    } else if(inputString[0] == 's'){  //set Data
          size_t delimiter = inputString.lastIndexOf(":");
          uint16_t datenNummer = inputString.substring(2,delimiter).toInt();
          switch(datenNummer){
            case DATA_ALL:  storeNewSettings(); break;
          }

          
          //den String splitten
    /*
          switch(datenNummer){
            case DATA_TEMP:                    break;
            case DATA_HYSTERESE:               break;
            case DATA_BATTERIE_TYP:            break;
            case DATA_BATTERIE_ZELLEN:         break;
            case DATA_BATTERIE_KAPAZITAET:     break;
            case DATA_BATTERIE_ABSCHALTGRENZE: break;
            case DATA_BATTERIE_SOLLWARNEN:     break;
            case DATA_BATTERIE_WARNSCHWELLE:   break;
            case DATA_BATTERIE_SOLLSPEICHERN:  break;
            case DATA_HEIZUNG_LAUFDAUER:       break;
            case DATA_HEIZUNG_PAUSEDAUER:      break;
            case DATA_HEIZUNG_PWMBOOST:        break;
            case DATA_VENTILATOR_LAUFDAUER:    break;
            case DATA_VENTILATOR_PAUSEDAUER:   break;
            case DATA_KALIBRIERUNG:            break;
            case DATA_LEDHELLIGKEIT:           break;
            case DATA_SAVE:                    break;
            case DATA_LOAD_DEFAULT:            break;
            case DATA_LOAD_PRE:                break;
          }
        */
    }
     // clear the string:
    inputString = "";
    stringComplete = false; 
}



void storeNewSettings() {
  
    size_t delimiter = inputString.lastIndexOf(":");
    uint16_t datenNummer = inputString.substring(2,delimiter).toInt();
    uint16_t settingsSize = sizeof(hbsettings);
    char pBuffer[settingsSize];
    uint16_t bufferIndex=0;
    
   for(int i = delimiter+1; i < inputString.length(); i+=2){
     byte sendb = getVal(inputString[i+1]) + (getVal(inputString[i]) << 4);
     pBuffer[bufferIndex++] = sendb;
   }
    memcpy(&hbsettings, &pBuffer, settingsSize);
    
}

byte getVal(char c){
   if(c >= '0' && c <= '9')
     return (byte)(c - '0');
   else
     return (byte)(c-'A'+10);
}

#endif //ENABLEPCCOM
